#include "animation_state_machine_editor.h"

#include "core/io/resource_loader.h"
#include "core/project_settings.h"
#include "math/delaunay.h"
#include "os/input.h"
#include "os/keyboard.h"
#include "scene/animation/animation_blend_tree.h"
#include "scene/animation/animation_player.h"
#include "scene/gui/menu_button.h"
#include "scene/gui/panel.h"
#include "scene/main/viewport.h"

void AnimationNodeStateMachineEditor::edit(AnimationNodeStateMachine *p_state_machine) {

	if (state_machine.is_valid()) {
		state_machine->disconnect("removed_from_graph", this, "_removed_from_graph");
	}

	if (p_state_machine) {
		state_machine = Ref<AnimationNodeStateMachine>(p_state_machine);
	} else {
		state_machine.unref();
	}

	if (state_machine.is_null()) {
		hide();
	} else {
		state_machine->connect("removed_from_graph", this, "_removed_from_graph");

		selected_transition_from = StringName();
		selected_transition_to = StringName();
		selected_node = StringName();
		_update_mode();
		_update_graph();
	}
}

void AnimationNodeStateMachineEditor::_state_machine_gui_input(const Ref<InputEvent> &p_event) {

	Ref<InputEventKey> k = p_event;
	if (tool_select->is_pressed() && k.is_valid() && k->is_pressed() && k->get_scancode() == KEY_DELETE && !k->is_echo()) {
		if (selected_node != StringName() || selected_transition_to != StringName() || selected_transition_from != StringName()) {
			_erase_selected();
			accept_event();
		}
	}

	Ref<InputEventMouseButton> mb = p_event;

	//Add new node
	if (mb.is_valid() && mb->is_pressed() && ((tool_select->is_pressed() && mb->get_button_index() == BUTTON_RIGHT) || (tool_create->is_pressed() && mb->get_button_index() == BUTTON_LEFT))) {
		menu->clear();
		animations_menu->clear();
		animations_to_add.clear();
		List<StringName> classes;
		classes.sort_custom<StringName::AlphCompare>();

		ClassDB::get_inheriters_from_class("AnimationRootNode", &classes);
		menu->add_submenu_item(TTR("Add Animation"), "animations");

		AnimationTree *gp = state_machine->get_tree();
		ERR_FAIL_COND(!gp);
		if (gp && gp->has_node(gp->get_animation_player())) {
			AnimationPlayer *ap = Object::cast_to<AnimationPlayer>(gp->get_node(gp->get_animation_player()));
			if (ap) {
				List<StringName> names;
				ap->get_animation_list(&names);
				for (List<StringName>::Element *E = names.front(); E; E = E->next()) {
					animations_menu->add_icon_item(get_icon("Animation", "EditorIcons"), E->get());
					animations_to_add.push_back(E->get());
				}
			}
		}

		for (List<StringName>::Element *E = classes.front(); E; E = E->next()) {

			String name = String(E->get()).replace_first("AnimationNode", "");
			if (name == "Animation")
				continue; // nope
			int idx = menu->get_item_count();
			menu->add_item(vformat("Add %s", name));
			menu->set_item_metadata(idx, E->get());
		}

		menu->set_global_position(state_machine_draw->get_global_transform().xform(mb->get_position()));
		menu->popup();
		add_node_pos = mb->get_position() / EDSCALE + state_machine->get_graph_offset();
	}

	// select node or push a field inside
	if (mb.is_valid() && !mb->get_shift() && mb->is_pressed() && tool_select->is_pressed() && mb->get_button_index() == BUTTON_LEFT) {

		selected_transition_from = StringName();
		selected_transition_to = StringName();
		selected_node = StringName();

		for (int i = node_rects.size() - 1; i >= 0; i--) { //inverse to draw order

			if (node_rects[i].play.has_point(mb->get_position())) { //edit name
				if (play_mode->get_selected() == 1 || !state_machine->is_playing()) {
					//start
					state_machine->start(node_rects[i].node_name);
				} else {
					//travel
					if (!state_machine->travel(node_rects[i].node_name)) {

						state_machine->start(node_rects[i].node_name);
						//removing this due to usability..
						//error_time = 5;
						//error_text = vformat(TTR("No path found from '%s' to '%s'."), state_machine->get_current_node(), node_rects[i].node_name);
					}
				}
				state_machine_draw->update();
				return;
			}

			if (node_rects[i].name.has_point(mb->get_position())) { //edit name

				Ref<StyleBox> line_sb = get_stylebox("normal", "LineEdit");

				Rect2 edit_rect = node_rects[i].name;
				edit_rect.position -= line_sb->get_offset();
				edit_rect.size += line_sb->get_minimum_size();

				name_edit->set_global_position(state_machine_draw->get_global_transform().xform(edit_rect.position));
				name_edit->set_size(edit_rect.size);
				name_edit->set_text(node_rects[i].node_name);
				name_edit->show_modal();
				name_edit->grab_focus();
				name_edit->select_all();

				prev_name = node_rects[i].node_name;
				return;
			}

			if (node_rects[i].edit.has_point(mb->get_position())) { //edit name
				call_deferred("_open_editor", node_rects[i].node_name);
				return;
			}

			if (node_rects[i].node.has_point(mb->get_position())) { //select node since nothing else was selected
				selected_node = node_rects[i].node_name;

				Ref<AnimationNode> anode = state_machine->get_node(selected_node);
				EditorNode::get_singleton()->push_item(anode.ptr(), "", true);
				state_machine_draw->update();
				dragging_selected_attempt = true;
				dragging_selected = false;
				drag_from = mb->get_position();
				snap_x = StringName();
				snap_y = StringName();
				_update_mode();
				return;
			}
		}

		//test the lines now
		int closest = -1;
		float closest_d = 1e20;
		for (int i = 0; i < transition_lines.size(); i++) {

			Vector2 s[2] = {
				transition_lines[i].from,
				transition_lines[i].to
			};
			Vector2 cpoint = Geometry::get_closest_point_to_segment_2d(mb->get_position(), s);
			float d = cpoint.distance_to(mb->get_position());
			if (d > transition_lines[i].width) {
				continue;
			}

			if (d < closest_d) {
				closest = i;
				closest_d = d;
			}
		}

		if (closest >= 0) {
			selected_transition_from = transition_lines[closest].from_node;
			selected_transition_to = transition_lines[closest].to_node;

			Ref<AnimationNodeStateMachineTransition> tr = state_machine->get_transition(closest);
			EditorNode::get_singleton()->push_item(tr.ptr(), "", true);
		}

		state_machine_draw->update();
		_update_mode();
	}

	//end moving node
	if (mb.is_valid() && dragging_selected_attempt && mb->get_button_index() == BUTTON_LEFT && !mb->is_pressed()) {

		if (dragging_selected) {

			Ref<AnimationNode> an = state_machine->get_node(selected_node);
			updating = true;
			undo_redo->create_action("Move Node");
			undo_redo->add_do_method(an.ptr(), "set_position", an->get_position() + drag_ofs / EDSCALE);
			undo_redo->add_undo_method(an.ptr(), "set_position", an->get_position());
			undo_redo->add_do_method(this, "_update_graph");
			undo_redo->add_undo_method(this, "_update_graph");
			undo_redo->commit_action();
			updating = false;
		}
		snap_x = StringName();
		snap_y = StringName();

		dragging_selected_attempt = false;
		dragging_selected = false;
		state_machine_draw->update();
	}

	//connect nodes
	if (mb.is_valid() && ((tool_select->is_pressed() && mb->get_shift()) || tool_connect->is_pressed()) && mb->get_button_index() == BUTTON_LEFT && mb->is_pressed()) {

		for (int i = node_rects.size() - 1; i >= 0; i--) { //inverse to draw order
			if (node_rects[i].node.has_point(mb->get_position())) { //select node since nothing else was selected
				connecting = true;
				connecting_from = node_rects[i].node_name;
				connecting_to = mb->get_position();
				connecting_to_node = StringName();
				return;
			}
		}
	}

	//end connecting nodes
	if (mb.is_valid() && connecting && mb->get_button_index() == BUTTON_LEFT && !mb->is_pressed()) {

		if (connecting_to_node != StringName()) {

			if (state_machine->has_transition(connecting_from, connecting_to_node)) {
				EditorNode::get_singleton()->show_warning("Transition exists!");

			} else {

				Ref<AnimationNodeStateMachineTransition> tr;
				tr.instance();
				tr->set_switch_mode(AnimationNodeStateMachineTransition::SwitchMode(transition_mode->get_selected()));

				updating = true;
				undo_redo->create_action("Add Transition");
				undo_redo->add_do_method(state_machine.ptr(), "add_transition", connecting_from, connecting_to_node, tr);
				undo_redo->add_undo_method(state_machine.ptr(), "remove_transition", connecting_from, connecting_to_node);
				undo_redo->add_do_method(this, "_update_graph");
				undo_redo->add_undo_method(this, "_update_graph");
				undo_redo->commit_action();
				updating = false;

				selected_transition_from = connecting_from;
				selected_transition_to = connecting_to_node;

				EditorNode::get_singleton()->push_item(tr.ptr(), "", true);
				_update_mode();
			}
		}
		connecting_to_node = StringName();
		connecting = false;
		state_machine_draw->update();
	}

	Ref<InputEventMouseMotion> mm = p_event;

	//pan window
	if (mm.is_valid() && mm->get_button_mask() & BUTTON_MASK_MIDDLE) {

		h_scroll->set_value(h_scroll->get_value() - mm->get_relative().x);
		v_scroll->set_value(v_scroll->get_value() - mm->get_relative().y);
	}

	//move mouse while connecting
	if (mm.is_valid() && connecting) {

		connecting_to = mm->get_position();
		connecting_to_node = StringName();
		state_machine_draw->update();

		for (int i = node_rects.size() - 1; i >= 0; i--) { //inverse to draw order
			if (node_rects[i].node_name != connecting_from && node_rects[i].node.has_point(connecting_to)) { //select node since nothing else was selected
				connecting_to_node = node_rects[i].node_name;
				return;
			}
		}
	}

	//move mouse while moving a node
	if (mm.is_valid() && dragging_selected_attempt) {

		dragging_selected = true;
		drag_ofs = mm->get_position() - drag_from;
		snap_x = StringName();
		snap_y = StringName();
		{
			//snap
			Vector2 cpos = state_machine->get_node(selected_node)->get_position() + drag_ofs / EDSCALE;
			List<StringName> nodes;
			state_machine->get_node_list(&nodes);

			float best_d_x = 1e20;
			float best_d_y = 1e20;

			for (List<StringName>::Element *E = nodes.front(); E; E = E->next()) {
				if (E->get() == selected_node)
					continue;
				Vector2 npos = state_machine->get_node(E->get())->get_position();

				float d_x = ABS(npos.x - cpos.x);
				if (d_x < MIN(5, best_d_x)) {
					drag_ofs.x -= cpos.x - npos.x;
					best_d_x = d_x;
					snap_x = E->get();
				}

				float d_y = ABS(npos.y - cpos.y);
				if (d_y < MIN(5, best_d_y)) {
					drag_ofs.y -= cpos.y - npos.y;
					best_d_y = d_y;
					snap_y = E->get();
				}
			}
		}

		state_machine_draw->update();
	}

	//put ibeam (text cursor) over names to make it clearer that they are editable
	if (mm.is_valid()) {

		state_machine_draw->grab_focus();

		bool over_text_now = false;
		String new_over_node = StringName();
		int new_over_node_what = -1;
		if (tool_select->is_pressed()) {

			for (int i = node_rects.size() - 1; i >= 0; i--) { //inverse to draw order

				if (node_rects[i].name.has_point(mm->get_position())) {
					over_text_now = true;
					break;
				}

				if (node_rects[i].node.has_point(mm->get_position())) {
					new_over_node = node_rects[i].node_name;
					if (node_rects[i].play.has_point(mm->get_position())) {
						new_over_node_what = 0;
					}
					if (node_rects[i].edit.has_point(mm->get_position())) {
						new_over_node_what = 1;
					}
				}
			}
		}

		if (new_over_node != over_node || new_over_node_what != over_node_what) {
			over_node = new_over_node;
			over_node_what = new_over_node_what;
			state_machine_draw->update();
		}

		if (over_text != over_text_now) {

			if (over_text_now) {
				state_machine_draw->set_default_cursor_shape(CURSOR_IBEAM);
			} else {
				state_machine_draw->set_default_cursor_shape(CURSOR_ARROW);
			}

			over_text = over_text_now;
		}
	}
}

void AnimationNodeStateMachineEditor::_add_menu_type(int p_index) {

	String type = menu->get_item_metadata(p_index);

	Object *obj = ClassDB::instance(type);
	ERR_FAIL_COND(!obj);
	AnimationNode *an = Object::cast_to<AnimationNode>(obj);
	ERR_FAIL_COND(!an);

	Ref<AnimationNode> node(an);
	node->set_position(add_node_pos);

	String base_name = type.replace_first("AnimationNode", "");
	int base = 1;
	String name = base_name;
	while (state_machine->has_node(name)) {
		base++;
		name = base_name + " " + itos(base);
	}

	updating = true;
	undo_redo->create_action("Add Node");
	undo_redo->add_do_method(state_machine.ptr(), "add_node", name, node);
	undo_redo->add_undo_method(state_machine.ptr(), "remove_node", name);
	undo_redo->add_do_method(this, "_update_graph");
	undo_redo->add_undo_method(this, "_update_graph");
	undo_redo->commit_action();
	updating = false;

	state_machine_draw->update();
}

void AnimationNodeStateMachineEditor::_add_animation_type(int p_index) {

	Ref<AnimationNodeAnimation> anim;
	anim.instance();

	anim->set_animation(animations_to_add[p_index]);

	String base_name = animations_to_add[p_index];
	int base = 1;
	String name = base_name;
	while (state_machine->has_node(name)) {
		base++;
		name = base_name + " " + itos(base);
	}

	anim->set_position(add_node_pos);

	updating = true;
	undo_redo->create_action("Add Node");
	undo_redo->add_do_method(state_machine.ptr(), "add_node", name, anim);
	undo_redo->add_undo_method(state_machine.ptr(), "remove_node", name);
	undo_redo->add_do_method(this, "_update_graph");
	undo_redo->add_undo_method(this, "_update_graph");
	undo_redo->commit_action();
	updating = false;

	state_machine_draw->update();
}

void AnimationNodeStateMachineEditor::_connection_draw(const Vector2 &p_from, const Vector2 &p_to, AnimationNodeStateMachineTransition::SwitchMode p_mode, bool p_enabled, bool p_selected, bool p_travel, bool p_auto_advance) {

	Color linecolor = get_color("font_color", "Label");
	Color icon_color(1, 1, 1);
	Color accent = get_color("accent_color", "Editor");

	if (!p_enabled) {
		linecolor.a *= 0.2;
		icon_color.a *= 0.2;
		accent.a *= 0.6;
	}

	Ref<Texture> icons[6] = {
		get_icon("TransitionImmediateBig", "EditorIcons"),
		get_icon("TransitionSyncBig", "EditorIcons"),
		get_icon("TransitionEndBig", "EditorIcons"),
		get_icon("TransitionImmediateAutoBig", "EditorIcons"),
		get_icon("TransitionSyncAutoBig", "EditorIcons"),
		get_icon("TransitionEndAutoBig", "EditorIcons")
	};

	if (p_selected) {
		state_machine_draw->draw_line(p_from, p_to, accent, 6, true);
	}

	if (p_travel) {
		linecolor = accent;
		linecolor.set_hsv(1.0, linecolor.get_s(), linecolor.get_v());
	}
	state_machine_draw->draw_line(p_from, p_to, linecolor, 2, true);

	Ref<Texture> icon = icons[p_mode + (p_auto_advance ? 3 : 0)];

	Transform2D xf;
	xf.elements[0] = (p_to - p_from).normalized();
	xf.elements[1] = xf.elements[0].tangent();
	xf.elements[2] = (p_from + p_to) * 0.5 - xf.elements[1] * icon->get_height() * 0.5 - xf.elements[0] * icon->get_height() * 0.5;

	state_machine_draw->draw_set_transform_matrix(xf);
	state_machine_draw->draw_texture(icon, Vector2(), icon_color);
	state_machine_draw->draw_set_transform_matrix(Transform2D());
}

void AnimationNodeStateMachineEditor::_clip_src_line_to_rect(Vector2 &r_from, Vector2 &r_to, const Rect2 &p_rect) {

	if (r_to == r_from)
		return;

	//this could be optimized...
	Vector2 n = (r_to - r_from).normalized();
	while (p_rect.has_point(r_from)) {
		r_from += n;
	}
}

void AnimationNodeStateMachineEditor::_clip_dst_line_to_rect(Vector2 &r_from, Vector2 &r_to, const Rect2 &p_rect) {

	if (r_to == r_from)
		return;

	//this could be optimized...
	Vector2 n = (r_to - r_from).normalized();
	while (p_rect.has_point(r_to)) {
		r_to -= n;
	}
}

void AnimationNodeStateMachineEditor::_state_machine_draw() {

	Ref<StyleBox> style = get_stylebox("frame", "GraphNode");
	Ref<StyleBox> style_selected = get_stylebox("selectedframe", "GraphNode");

	Ref<Font> font = get_font("title_font", "GraphNode");
	Color font_color = get_color("title_color", "GraphNode");
	Ref<Texture> play = get_icon("Play", "EditorIcons");
	Ref<Texture> auto_play = get_icon("AutoPlay", "EditorIcons");
	Ref<Texture> edit = get_icon("Edit", "EditorIcons");
	Color accent = get_color("accent_color", "Editor");
	Color linecolor = get_color("font_color", "Label");
	linecolor.a *= 0.3;
	Ref<StyleBox> playing_overlay = get_stylebox("position", "GraphNode");

	bool playing = state_machine->is_playing();
	StringName current = state_machine->get_current_node();
	StringName blend_from = state_machine->get_blend_from_node();
	Vector<StringName> travel_path = state_machine->get_travel_path();

	if (state_machine_draw->has_focus()) {
		state_machine_draw->draw_rect(Rect2(Point2(), state_machine_draw->get_size()), accent, false);
	}
	int sep = 3 * EDSCALE;

	List<StringName> nodes;
	state_machine->get_node_list(&nodes);

	node_rects.clear();
	Rect2 scroll_range(Point2(), state_machine_draw->get_size());

	//snap lines
	if (dragging_selected) {

		Vector2 from = (state_machine->get_node(selected_node)->get_position() * EDSCALE) + drag_ofs - state_machine->get_graph_offset() * EDSCALE;
		if (snap_x != StringName()) {
			Vector2 to = (state_machine->get_node(snap_x)->get_position() * EDSCALE) - state_machine->get_graph_offset() * EDSCALE;
			state_machine_draw->draw_line(from, to, linecolor, 2);
		}
		if (snap_y != StringName()) {
			Vector2 to = (state_machine->get_node(snap_y)->get_position() * EDSCALE) - state_machine->get_graph_offset() * EDSCALE;
			state_machine_draw->draw_line(from, to, linecolor, 2);
		}
	}

	//pre pass nodes so we know the rectangles
	for (List<StringName>::Element *E = nodes.front(); E; E = E->next()) {

		Ref<AnimationNode> anode = state_machine->get_node(E->get());
		String name = E->get();
		bool needs_editor = EditorNode::get_singleton()->item_has_editor(anode.ptr());
		Ref<StyleBox> sb = E->get() == selected_node ? style_selected : style;

		Size2 s = sb->get_minimum_size();
		int strsize = font->get_string_size(name).width;
		s.width += strsize;
		s.height += MAX(font->get_height(), play->get_height());
		s.width += sep + play->get_width();
		if (needs_editor) {
			s.width += sep + edit->get_width();
		}

		Vector2 offset;
		offset += anode->get_position() * EDSCALE;
		if (selected_node == E->get() && dragging_selected) {
			offset += drag_ofs;
		}
		offset -= s / 2;
		offset = offset.floor();

		//prepre rect

		NodeRect nr;
		nr.node = Rect2(offset, s);
		nr.node_name = E->get();

		scroll_range = scroll_range.merge(nr.node); //merge with range

		//now scroll it to draw
		nr.node.position -= state_machine->get_graph_offset() * EDSCALE;

		node_rects.push_back(nr);
	}

	transition_lines.clear();

	//draw conecting line for potential new transition
	if (connecting) {
		Vector2 from = (state_machine->get_node(connecting_from)->get_position() * EDSCALE) - state_machine->get_graph_offset() * EDSCALE;
		Vector2 to;
		if (connecting_to_node != StringName()) {
			to = (state_machine->get_node(connecting_to_node)->get_position() * EDSCALE) - state_machine->get_graph_offset() * EDSCALE;
		} else {
			to = connecting_to;
		}

		for (int i = 0; i < node_rects.size(); i++) {
			if (node_rects[i].node_name == connecting_from) {
				_clip_src_line_to_rect(from, to, node_rects[i].node);
			}
			if (node_rects[i].node_name == connecting_to_node) {
				_clip_dst_line_to_rect(from, to, node_rects[i].node);
			}
		}

		_connection_draw(from, to, AnimationNodeStateMachineTransition::SwitchMode(transition_mode->get_selected()), true, false, false, false);
	}

	Ref<Texture> tr_reference_icon = get_icon("TransitionImmediateBig", "EditorIcons");
	float tr_bidi_offset = int(tr_reference_icon->get_height() * 0.8);

	//draw transition lines
	for (int i = 0; i < state_machine->get_transition_count(); i++) {

		TransitionLine tl;
		tl.from_node = state_machine->get_transition_from(i);
		Vector2 ofs_from = (dragging_selected && tl.from_node == selected_node) ? drag_ofs : Vector2();
		tl.from = (state_machine->get_node(tl.from_node)->get_position() * EDSCALE) + ofs_from - state_machine->get_graph_offset() * EDSCALE;

		tl.to_node = state_machine->get_transition_to(i);
		Vector2 ofs_to = (dragging_selected && tl.to_node == selected_node) ? drag_ofs : Vector2();
		tl.to = (state_machine->get_node(tl.to_node)->get_position() * EDSCALE) + ofs_to - state_machine->get_graph_offset() * EDSCALE;

		Ref<AnimationNodeStateMachineTransition> tr = state_machine->get_transition(i);
		tl.disabled = tr->is_disabled();
		tl.auto_advance = tr->has_auto_advance();
		tl.mode = tr->get_switch_mode();
		tl.width = tr_bidi_offset;

		if (state_machine->has_transition(tl.to_node, tl.from_node)) { //offset if same exists
			Vector2 offset = -(tl.from - tl.to).normalized().tangent() * tr_bidi_offset;
			tl.from += offset;
			tl.to += offset;
		}

		for (int i = 0; i < node_rects.size(); i++) {
			if (node_rects[i].node_name == tl.from_node) {
				_clip_src_line_to_rect(tl.from, tl.to, node_rects[i].node);
			}
			if (node_rects[i].node_name == tl.to_node) {
				_clip_dst_line_to_rect(tl.from, tl.to, node_rects[i].node);
			}
		}

		bool selected = selected_transition_from == tl.from_node && selected_transition_to == tl.to_node;

		bool travel = false;

		if (blend_from == tl.from_node && current == tl.to_node) {
			travel = true;
		}

		if (travel_path.size()) {

			if (current == tl.from_node && travel_path[0] == tl.to_node) {
				travel = true;
			} else {
				for (int j = 0; j < travel_path.size() - 1; j++) {
					if (travel_path[j] == tl.from_node && travel_path[j + 1] == tl.to_node) {
						travel = true;
						break;
					}
				}
			}
		}
		_connection_draw(tl.from, tl.to, tl.mode, !tl.disabled, selected, travel, tl.auto_advance);

		transition_lines.push_back(tl);
	}

	//draw actual nodes
	for (int i = 0; i < node_rects.size(); i++) {

		String name = node_rects[i].node_name;
		Ref<AnimationNode> anode = state_machine->get_node(name);
		bool needs_editor = EditorNode::get_singleton()->item_has_editor(anode.ptr());
		Ref<StyleBox> sb = name == selected_node ? style_selected : style;
		int strsize = font->get_string_size(name).width;

		NodeRect &nr = node_rects[i];

		Vector2 offset = nr.node.position;
		int h = nr.node.size.height;

		//prepre rect

		//now scroll it to draw
		state_machine_draw->draw_style_box(sb, nr.node);

		if (playing && (blend_from == name || current == name || travel_path.find(name) != -1)) {
			state_machine_draw->draw_style_box(playing_overlay, nr.node);
		}

		bool onstart = state_machine->get_start_node() == name;
		if (onstart) {
			state_machine_draw->draw_string(font, offset + Vector2(0, -font->get_height() - 3 * EDSCALE + font->get_ascent()), TTR("Start"), font_color);
		}

		if (state_machine->get_end_node() == name) {

			int endofs = nr.node.size.x - font->get_string_size(TTR("End")).x;
			state_machine_draw->draw_string(font, offset + Vector2(endofs, -font->get_height() - 3 * EDSCALE + font->get_ascent()), TTR("End"), font_color);
		}

		offset.x += sb->get_offset().x;

		nr.play.position = offset + Vector2(0, (h - play->get_height()) / 2).floor();
		nr.play.size = play->get_size();

		Ref<Texture> play_tex = onstart ? auto_play : play;

		if (over_node == name && over_node_what == 0) {
			state_machine_draw->draw_texture(play_tex, nr.play.position, accent);
		} else {
			state_machine_draw->draw_texture(play_tex, nr.play.position);
		}
		offset.x += sep + play->get_width();

		nr.name.position = offset + Vector2(0, (h - font->get_height()) / 2).floor();
		nr.name.size = Vector2(strsize, font->get_height());

		state_machine_draw->draw_string(font, nr.name.position + Vector2(0, font->get_ascent()), name, font_color);
		offset.x += strsize + sep;

		if (needs_editor) {
			nr.edit.position = offset + Vector2(0, (h - edit->get_height()) / 2).floor();
			nr.edit.size = edit->get_size();

			if (over_node == name && over_node_what == 1) {
				state_machine_draw->draw_texture(edit, nr.edit.position, accent);
			} else {
				state_machine_draw->draw_texture(edit, nr.edit.position);
			}
			offset.x += sep + edit->get_width();
		}
	}

	scroll_range = scroll_range.grow(200 * EDSCALE);

	//adjust scrollbars
	updating = true;
	h_scroll->set_min(scroll_range.position.x);
	h_scroll->set_max(scroll_range.position.x + scroll_range.size.x);
	h_scroll->set_page(state_machine_draw->get_size().x);
	h_scroll->set_value(state_machine->get_graph_offset().x);

	v_scroll->set_min(scroll_range.position.y);
	v_scroll->set_max(scroll_range.position.y + scroll_range.size.y);
	v_scroll->set_page(state_machine_draw->get_size().y);
	v_scroll->set_value(state_machine->get_graph_offset().y);
	updating = false;

	state_machine_play_pos->update();
}

void AnimationNodeStateMachineEditor::_state_machine_pos_draw() {

	if (!state_machine->is_playing())
		return;

	int idx = -1;
	for (int i = 0; node_rects.size(); i++) {
		if (node_rects[i].node_name == state_machine->get_current_node()) {
			idx = i;
			break;
		}
	}

	if (idx == -1)
		return;

	NodeRect &nr = node_rects[idx];

	Vector2 from;
	from.x = nr.play.position.x;
	from.y = (nr.play.position.y + nr.play.size.y + nr.node.position.y + nr.node.size.y) * 0.5;

	Vector2 to;
	if (nr.edit.size.x) {
		to.x = nr.edit.position.x + nr.edit.size.x;
	} else {
		to.x = nr.name.position.x + nr.name.size.x;
	}
	to.y = from.y;

	float len = MAX(0.0001, state_machine->get_current_length());

	float pos = CLAMP(state_machine->get_current_play_pos(), 0, len);
	float c = pos / len;
	Color fg = get_color("font_color", "Label");
	Color bg = fg;
	bg.a *= 0.3;

	state_machine_play_pos->draw_line(from, to, bg, 2);

	to = from.linear_interpolate(to, c);

	state_machine_play_pos->draw_line(from, to, fg, 2);
}

void AnimationNodeStateMachineEditor::_update_graph() {

	if (updating)
		return;

	updating = true;

	if (state_machine->get_parent().is_valid()) {
		goto_parent_hbox->show();
	} else {
		goto_parent_hbox->hide();
	}

	state_machine_draw->update();

	updating = false;
}

void AnimationNodeStateMachineEditor::_notification(int p_what) {

	if (p_what == NOTIFICATION_ENTER_TREE || p_what == NOTIFICATION_THEME_CHANGED) {
		error_panel->add_style_override("panel", get_stylebox("bg", "Tree"));
		error_label->add_color_override("font_color", get_color("error_color", "Editor"));
		panel->add_style_override("panel", get_stylebox("bg", "Tree"));
		goto_parent->set_icon(get_icon("MoveUp", "EditorIcons"));

		tool_select->set_icon(get_icon("ToolSelect", "EditorIcons"));
		tool_create->set_icon(get_icon("ToolAddNode", "EditorIcons"));
		tool_connect->set_icon(get_icon("ToolConnect", "EditorIcons"));

		transition_mode->clear();
		transition_mode->add_icon_item(get_icon("TransitionImmediate", "EditorIcons"), TTR("Immediate"));
		transition_mode->add_icon_item(get_icon("TransitionSync", "EditorIcons"), TTR("Sync"));
		transition_mode->add_icon_item(get_icon("TransitionEnd", "EditorIcons"), TTR("At End"));

		//force filter on those, so they deform better
		get_icon("TransitionImmediateBig", "EditorIcons")->set_flags(Texture::FLAG_FILTER);
		get_icon("TransitionEndBig", "EditorIcons")->set_flags(Texture::FLAG_FILTER);
		get_icon("TransitionSyncBig", "EditorIcons")->set_flags(Texture::FLAG_FILTER);
		get_icon("TransitionImmediateAutoBig", "EditorIcons")->set_flags(Texture::FLAG_FILTER);
		get_icon("TransitionEndAutoBig", "EditorIcons")->set_flags(Texture::FLAG_FILTER);
		get_icon("TransitionSyncAutoBig", "EditorIcons")->set_flags(Texture::FLAG_FILTER);

		tool_erase->set_icon(get_icon("Remove", "EditorIcons"));
		tool_autoplay->set_icon(get_icon("AutoPlay", "EditorIcons"));
		tool_end->set_icon(get_icon("AutoEnd", "EditorIcons"));

		play_mode->clear();
		play_mode->add_icon_item(get_icon("PlayTravel", "EditorIcons"), TTR("Travel"));
		play_mode->add_icon_item(get_icon("Play", "EditorIcons"), TTR("Immediate"));
	}

	if (p_what == NOTIFICATION_PROCESS) {

		String error;

		if (error_time > 0) {
			error = error_text;
			error_time -= get_process_delta_time();
		} else if (!state_machine->get_tree()) {
			error = TTR("StateMachine does not belong to an AnimationTree node.");
		} else if (!state_machine->get_tree()->is_active()) {
			error = TTR("AnimationTree is inactive.\nActivate to enable playback, check node warnings if activation fails.");
		} else if (state_machine->get_tree()->is_state_invalid()) {
			error = state_machine->get_tree()->get_invalid_state_reason();
		} else if (state_machine->get_parent().is_valid() && state_machine->get_parent()->is_class("AnimationNodeStateMachine")) {
			if (state_machine->get_start_node() == StringName() || state_machine->get_end_node() == StringName()) {
				error = TTR("Start and end nodes are needed for a sub-transition.");
			}
		}

		if (error != error_label->get_text()) {
			error_label->set_text(error);
			if (error != String()) {
				error_panel->show();
			} else {
				error_panel->hide();
			}
		}

		for (int i = 0; i < transition_lines.size(); i++) {
			int tidx = -1;
			for (int j = 0; j < state_machine->get_transition_count(); j++) {
				if (transition_lines[i].from_node == state_machine->get_transition_from(j) && transition_lines[i].to_node == state_machine->get_transition_to(j)) {
					tidx = j;
					break;
				}
			}

			if (tidx == -1) { //missing transition, should redraw
				state_machine_draw->update();
				break;
			}

			if (transition_lines[i].disabled != state_machine->get_transition(tidx)->is_disabled()) {
				state_machine_draw->update();
				break;
			}

			if (transition_lines[i].auto_advance != state_machine->get_transition(tidx)->has_auto_advance()) {
				state_machine_draw->update();
				break;
			}

			if (transition_lines[i].mode != state_machine->get_transition(tidx)->get_switch_mode()) {
				state_machine_draw->update();
				break;
			}
		}

		bool same_travel_path = true;
		Vector<StringName> tp = state_machine->get_travel_path();

		{

			if (last_travel_path.size() != tp.size()) {
				same_travel_path = false;
			} else {
				for (int i = 0; i < last_travel_path.size(); i++) {
					if (last_travel_path[i] != tp[i]) {
						same_travel_path = false;
						break;
					}
				}
			}
		}

		//update if travel state changed
		if (!same_travel_path || last_active != state_machine->is_playing() || last_current_node != state_machine->get_current_node() || last_blend_from_node != state_machine->get_blend_from_node()) {

			state_machine_draw->update();
			last_travel_path = tp;
			last_current_node = state_machine->get_current_node();
			last_active = state_machine->is_playing();
			last_blend_from_node = state_machine->get_blend_from_node();
			state_machine_play_pos->update();
		}

		if (last_play_pos != state_machine->get_current_play_pos()) {

			last_play_pos = state_machine->get_current_play_pos();
			state_machine_play_pos->update();
		}
	}

	if (p_what == NOTIFICATION_VISIBILITY_CHANGED) {
		over_node = StringName();
	}
}

void AnimationNodeStateMachineEditor::_open_editor(const String &p_name) {
	Ref<AnimationNode> an = state_machine->get_node(p_name);
	ERR_FAIL_COND(!an.is_valid());
	EditorNode::get_singleton()->edit_item(an.ptr());
}

void AnimationNodeStateMachineEditor::_goto_parent() {

	EditorNode::get_singleton()->edit_item(state_machine->get_parent().ptr());
}

void AnimationNodeStateMachineEditor::_removed_from_graph() {
	EditorNode::get_singleton()->edit_item(NULL);
}

void AnimationNodeStateMachineEditor::_name_edited(const String &p_text) {

	String new_name = p_text;

	ERR_FAIL_COND(new_name == "" || new_name.find(".") != -1 || new_name.find("/") != -1)

	ERR_FAIL_COND(new_name == prev_name);

	String base_name = new_name;
	int base = 1;
	String name = base_name;
	while (state_machine->has_node(name)) {
		base++;
		name = base_name + " " + itos(base);
	}

	updating = true;
	undo_redo->create_action("Node Renamed");
	undo_redo->add_do_method(state_machine.ptr(), "rename_node", prev_name, name);
	undo_redo->add_undo_method(state_machine.ptr(), "rename_node", name, prev_name);
	undo_redo->add_do_method(this, "_update_graph");
	undo_redo->add_undo_method(this, "_update_graph");
	undo_redo->commit_action();
	updating = false;

	state_machine_draw->update();

	name_edit->hide();
}

void AnimationNodeStateMachineEditor::_scroll_changed(double) {
	if (updating)
		return;

	state_machine->set_graph_offset(Vector2(h_scroll->get_value(), v_scroll->get_value()));
	state_machine_draw->update();
}

void AnimationNodeStateMachineEditor::_erase_selected() {

	if (selected_node != StringName() && state_machine->has_node(selected_node)) {
		updating = true;
		undo_redo->create_action("Node Removed");
		undo_redo->add_do_method(state_machine.ptr(), "remove_node", selected_node);
		undo_redo->add_undo_method(state_machine.ptr(), "add_node", selected_node, state_machine->get_node(selected_node));
		for (int i = 0; i < state_machine->get_transition_count(); i++) {
			String from = state_machine->get_transition_from(i);
			String to = state_machine->get_transition_to(i);
			if (from == selected_node || to == selected_node) {
				undo_redo->add_undo_method(state_machine.ptr(), "add_transition", from, to, state_machine->get_transition(i));
			}
		}
		if (String(state_machine->get_start_node()) == selected_node) {
			undo_redo->add_undo_method(state_machine.ptr(), "set_start_node", selected_node);
		}
		undo_redo->add_do_method(this, "_update_graph");
		undo_redo->add_undo_method(this, "_update_graph");
		undo_redo->commit_action();
		updating = false;
		selected_node = StringName();
	}

	if (selected_transition_to != StringName() && selected_transition_from != StringName() && state_machine->has_transition(selected_transition_from, selected_transition_to)) {

		Ref<AnimationNodeStateMachineTransition> tr = state_machine->get_transition(state_machine->find_transition(selected_transition_from, selected_transition_to));
		updating = true;
		undo_redo->create_action("Transition Removed");
		undo_redo->add_do_method(state_machine.ptr(), "remove_transition", selected_transition_from, selected_transition_to);
		undo_redo->add_undo_method(state_machine.ptr(), "add_transition", selected_transition_from, selected_transition_to, tr);
		undo_redo->add_do_method(this, "_update_graph");
		undo_redo->add_undo_method(this, "_update_graph");
		undo_redo->commit_action();
		updating = false;
		selected_transition_from = StringName();
		selected_transition_to = StringName();
	}

	state_machine_draw->update();
}

void AnimationNodeStateMachineEditor::_autoplay_selected() {

	if (selected_node != StringName() && state_machine->has_node(selected_node)) {

		StringName new_start_node;
		if (state_machine->get_start_node() == selected_node) { //toggle it
			new_start_node = StringName();
		} else {
			new_start_node = selected_node;
		}

		updating = true;
		undo_redo->create_action("Set Start Node (Autoplay)");
		undo_redo->add_do_method(state_machine.ptr(), "set_start_node", new_start_node);
		undo_redo->add_undo_method(state_machine.ptr(), "set_start_node", state_machine->get_start_node());
		undo_redo->add_do_method(this, "_update_graph");
		undo_redo->add_undo_method(this, "_update_graph");
		undo_redo->commit_action();
		updating = false;
		state_machine_draw->update();
	}
}

void AnimationNodeStateMachineEditor::_end_selected() {

	if (selected_node != StringName() && state_machine->has_node(selected_node)) {

		StringName new_end_node;
		if (state_machine->get_end_node() == selected_node) { //toggle it
			new_end_node = StringName();
		} else {
			new_end_node = selected_node;
		}

		updating = true;
		undo_redo->create_action("Set Start Node (Autoplay)");
		undo_redo->add_do_method(state_machine.ptr(), "set_end_node", new_end_node);
		undo_redo->add_undo_method(state_machine.ptr(), "set_end_node", state_machine->get_end_node());
		undo_redo->add_do_method(this, "_update_graph");
		undo_redo->add_undo_method(this, "_update_graph");
		undo_redo->commit_action();
		updating = false;
		state_machine_draw->update();
	}
}
void AnimationNodeStateMachineEditor::_update_mode() {

	if (tool_select->is_pressed()) {
		tool_erase_hb->show();
		tool_erase->set_disabled(selected_node == StringName() && selected_transition_from == StringName() && selected_transition_to == StringName());
		tool_autoplay->set_disabled(selected_node == StringName());
		tool_end->set_disabled(selected_node == StringName());
	} else {
		tool_erase_hb->hide();
	}
}

void AnimationNodeStateMachineEditor::_bind_methods() {

	ClassDB::bind_method("_state_machine_gui_input", &AnimationNodeStateMachineEditor::_state_machine_gui_input);
	ClassDB::bind_method("_state_machine_draw", &AnimationNodeStateMachineEditor::_state_machine_draw);
	ClassDB::bind_method("_state_machine_pos_draw", &AnimationNodeStateMachineEditor::_state_machine_pos_draw);
	ClassDB::bind_method("_update_graph", &AnimationNodeStateMachineEditor::_update_graph);

	ClassDB::bind_method("_add_menu_type", &AnimationNodeStateMachineEditor::_add_menu_type);
	ClassDB::bind_method("_add_animation_type", &AnimationNodeStateMachineEditor::_add_animation_type);

	ClassDB::bind_method("_name_edited", &AnimationNodeStateMachineEditor::_name_edited);

	ClassDB::bind_method("_goto_parent", &AnimationNodeStateMachineEditor::_goto_parent);
	ClassDB::bind_method("_removed_from_graph", &AnimationNodeStateMachineEditor::_removed_from_graph);

	ClassDB::bind_method("_open_editor", &AnimationNodeStateMachineEditor::_open_editor);
	ClassDB::bind_method("_scroll_changed", &AnimationNodeStateMachineEditor::_scroll_changed);

	ClassDB::bind_method("_erase_selected", &AnimationNodeStateMachineEditor::_erase_selected);
	ClassDB::bind_method("_autoplay_selected", &AnimationNodeStateMachineEditor::_autoplay_selected);
	ClassDB::bind_method("_end_selected", &AnimationNodeStateMachineEditor::_end_selected);
	ClassDB::bind_method("_update_mode", &AnimationNodeStateMachineEditor::_update_mode);
}

AnimationNodeStateMachineEditor *AnimationNodeStateMachineEditor::singleton = NULL;

AnimationNodeStateMachineEditor::AnimationNodeStateMachineEditor() {

	singleton = this;
	updating = false;

	HBoxContainer *top_hb = memnew(HBoxContainer);
	add_child(top_hb);

	goto_parent_hbox = memnew(HBoxContainer);
	goto_parent = memnew(ToolButton);
	goto_parent->connect("pressed", this, "_goto_parent", varray(), CONNECT_DEFERRED);
	goto_parent_hbox->add_child(goto_parent);
	goto_parent_hbox->add_child(memnew(VSeparator));
	top_hb->add_child(goto_parent_hbox);

	Ref<ButtonGroup> bg;
	bg.instance();

	tool_select = memnew(ToolButton);
	top_hb->add_child(tool_select);
	tool_select->set_toggle_mode(true);
	tool_select->set_button_group(bg);
	tool_select->set_pressed(true);
	tool_select->set_tooltip(TTR("Select and move nodes.\nRMB to add new nodes.\nShift+LMB to create connections."));
	tool_select->connect("pressed", this, "_update_mode", varray(), CONNECT_DEFERRED);

	tool_create = memnew(ToolButton);
	top_hb->add_child(tool_create);
	tool_create->set_toggle_mode(true);
	tool_create->set_button_group(bg);
	tool_create->set_tooltip(TTR("Create new nodes."));
	tool_create->connect("pressed", this, "_update_mode", varray(), CONNECT_DEFERRED);

	tool_connect = memnew(ToolButton);
	top_hb->add_child(tool_connect);
	tool_connect->set_toggle_mode(true);
	tool_connect->set_button_group(bg);
	tool_connect->set_tooltip(TTR("Connect nodes."));
	tool_connect->connect("pressed", this, "_update_mode", varray(), CONNECT_DEFERRED);

	tool_erase_hb = memnew(HBoxContainer);
	top_hb->add_child(tool_erase_hb);
	tool_erase_hb->add_child(memnew(VSeparator));
	tool_erase = memnew(ToolButton);
	tool_erase->set_tooltip(TTR("Remove selected node or transition"));
	tool_erase_hb->add_child(tool_erase);
	tool_erase->connect("pressed", this, "_erase_selected");
	tool_erase->set_disabled(true);

	tool_erase_hb->add_child(memnew(VSeparator));

	tool_autoplay = memnew(ToolButton);
	tool_autoplay->set_tooltip(TTR("Toggle autoplay this animation on start, restart or seek to zero."));
	tool_erase_hb->add_child(tool_autoplay);
	tool_autoplay->connect("pressed", this, "_autoplay_selected");
	tool_autoplay->set_disabled(true);

	tool_end = memnew(ToolButton);
	tool_end->set_tooltip(TTR("Set the end animation. This is useful for sub-transitions."));
	tool_erase_hb->add_child(tool_end);
	tool_end->connect("pressed", this, "_end_selected");
	tool_end->set_disabled(true);

	top_hb->add_child(memnew(VSeparator));
	top_hb->add_child(memnew(Label(TTR("Transition: "))));
	transition_mode = memnew(OptionButton);
	top_hb->add_child(transition_mode);

	top_hb->add_spacer();

	top_hb->add_child(memnew(Label("Play Mode:")));
	play_mode = memnew(OptionButton);
	top_hb->add_child(play_mode);

	GridContainer *main_grid = memnew(GridContainer);
	main_grid->set_columns(2);
	add_child(main_grid);
	main_grid->set_v_size_flags(SIZE_EXPAND_FILL);

	panel = memnew(PanelContainer);
	panel->set_clip_contents(true);
	main_grid->add_child(panel);
	panel->set_h_size_flags(SIZE_EXPAND_FILL);

	state_machine_draw = memnew(Control);
	state_machine_draw->connect("gui_input", this, "_state_machine_gui_input");
	state_machine_draw->connect("draw", this, "_state_machine_draw");
	state_machine_draw->set_focus_mode(FOCUS_ALL);

	state_machine_play_pos = memnew(Control);
	state_machine_draw->add_child(state_machine_play_pos);
	state_machine_play_pos->set_mouse_filter(MOUSE_FILTER_PASS); //pass all to parent
	state_machine_play_pos->set_anchors_and_margins_preset(PRESET_WIDE);
	state_machine_play_pos->connect("draw", this, "_state_machine_pos_draw");

	panel->add_child(state_machine_draw);
	panel->set_v_size_flags(SIZE_EXPAND_FILL);

	v_scroll = memnew(VScrollBar);
	main_grid->add_child(v_scroll);
	v_scroll->connect("value_changed", this, "_scroll_changed");

	h_scroll = memnew(HScrollBar);
	main_grid->add_child(h_scroll);
	h_scroll->connect("value_changed", this, "_scroll_changed");

	main_grid->add_child(memnew(Control)); //empty bottom right

	error_panel = memnew(PanelContainer);
	add_child(error_panel);
	error_label = memnew(Label);
	error_panel->add_child(error_label);
	error_label->set_text("eh");

	undo_redo = EditorNode::get_singleton()->get_undo_redo();

	set_custom_minimum_size(Size2(0, 300 * EDSCALE));

	menu = memnew(PopupMenu);
	add_child(menu);
	menu->connect("index_pressed", this, "_add_menu_type");

	animations_menu = memnew(PopupMenu);
	menu->add_child(animations_menu);
	animations_menu->set_name("animations");
	animations_menu->connect("index_pressed", this, "_add_animation_type");

	name_edit = memnew(LineEdit);
	state_machine_draw->add_child(name_edit);
	name_edit->hide();
	name_edit->connect("text_entered", this, "_name_edited");
	name_edit->set_as_toplevel(true);

	over_text = false;

	over_node_what = -1;
	dragging_selected_attempt = false;
	connecting = false;

	last_active = false;

	error_time = 0;
}

void AnimationNodeStateMachineEditorPlugin::edit(Object *p_object) {

	anim_tree_editor->edit(Object::cast_to<AnimationNodeStateMachine>(p_object));
}

bool AnimationNodeStateMachineEditorPlugin::handles(Object *p_object) const {

	return p_object->is_class("AnimationNodeStateMachine");
}

void AnimationNodeStateMachineEditorPlugin::make_visible(bool p_visible) {

	if (p_visible) {
		//editor->hide_animation_player_editors();
		//editor->animation_panel_make_visible(true);
		button->show();
		editor->make_bottom_panel_item_visible(anim_tree_editor);
		anim_tree_editor->set_process(true);
	} else {

		if (anim_tree_editor->is_visible_in_tree())
			editor->hide_bottom_panel();
		button->hide();
		anim_tree_editor->set_process(false);
	}
}

AnimationNodeStateMachineEditorPlugin::AnimationNodeStateMachineEditorPlugin(EditorNode *p_node) {

	editor = p_node;
	anim_tree_editor = memnew(AnimationNodeStateMachineEditor);
	anim_tree_editor->set_custom_minimum_size(Size2(0, 300));

	button = editor->add_bottom_panel_item(TTR("StateMachine"), anim_tree_editor);
	button->hide();
}

AnimationNodeStateMachineEditorPlugin::~AnimationNodeStateMachineEditorPlugin() {
}
